@@ -17,7 +17,7 @@ functions --- functions that estimate the utility of evaluating sets of candidat
 """
 from __future__ import annotations
 
-from typing import Mapping, Optional, cast
+from typing import Callable, Mapping, Optional, cast
 
 import tensorflow as tf
 import tensorflow_probability as tfp
@@ -30,7 +30,7 @@ from ...models.interfaces import (
     SupportsReparamSamplerObservationNoise,
 )
 from ...space import SearchSpace
-from ...types import TensorType
+from ...types import Tag, TensorType
 from ...utils import DEFAULTS
 from ..interface import (
     AcquisitionFunction,
@@ -40,6 +40,56 @@ from ..interface import (
     SingleModelAcquisitionBuilder,
     SingleModelVectorizedAcquisitionBuilder,
 )
+from .utils import MultivariateNormalCDF
+
+
+class ProbabilityOfImprovement(SingleModelAcquisitionBuilder[ProbabilisticModel]):
+    """
+    Builder for the probability of improvement function, where the "best" value
+    is taken to be the minimum of the posterior mean at observed points.
+    """
+
+    def __repr__(self) -> str:
+        """"""
+        return "ProbabilityOfImprovement()"
+
+    def prepare_acquisition_function(
+        self, model: ProbabilisticModel, dataset: Optional[Dataset] = None
+    ) -> AcquisitionFunction:
+        """
+        :param model: The model.
+        :param dataset: The data from the observer. Must be populated.
+        :return: The probability of improvement function. This function will raise
+            :exc:`ValueError` or :exc:`~tf.errors.InvalidArgumentError` if used with a batch size
+            greater than one.
+        :raise tf.errors.InvalidArgumentError: If ``dataset`` is empty.
+        """
+        tf.debugging.Assert(dataset is not None, [])
+        dataset = cast(Dataset, dataset)
+        tf.debugging.assert_positive(len(dataset), message="Dataset must be populated.")
+        mean, _ = model.predict(dataset.query_points)
+        eta = tf.reduce_min(mean, axis=0)[0]
+        return probability_below_threshold(model, eta)
+
+    def update_acquisition_function(
+        self,
+        function: AcquisitionFunction,
+        model: ProbabilisticModel,
+        dataset: Optional[Dataset] = None,
+    ) -> AcquisitionFunction:
+        """
+        :param function: The acquisition function to update.
+        :param model: The model.
+        :param dataset: The data from the observer.  Must be populated.
+        """
+        tf.debugging.Assert(dataset is not None, [])
+        dataset = cast(Dataset, dataset)
+        tf.debugging.assert_positive(len(dataset), message="Dataset must be populated.")
+        tf.debugging.Assert(isinstance(function, probability_below_threshold), [tf.constant([])])
+        mean, _ = model.predict(dataset.query_points)
+        eta = tf.reduce_min(mean, axis=0)[0]
+        function.update(eta)  # type: ignore
+        return function
 
 
 class ExpectedImprovement(SingleModelAcquisitionBuilder[ProbabilisticModel]):
@@ -329,8 +379,8 @@ def lower_confidence_bound(model: ProbabilisticModel, beta: float) -> Acquisitio
 
 class ProbabilityOfFeasibility(SingleModelAcquisitionBuilder[ProbabilisticModel]):
     r"""
-    Builder for the :func:`probability_of_feasibility` acquisition function, defined in
-    :cite:`gardner14` as
+    Uses the :func:`probability_below_threshold` function to build a
+    probability of feasiblity acquisition function, defined in :cite:`gardner14` as
 
     .. math::
 
@@ -371,7 +421,7 @@ class ProbabilityOfFeasibility(SingleModelAcquisitionBuilder[ProbabilisticModel]
             :exc:`ValueError` or :exc:`~tf.errors.InvalidArgumentError` if used with a batch size
             greater than one.
         """
-        return probability_of_feasibility(model, self.threshold)
+        return probability_below_threshold(model, self.threshold)
 
     def update_acquisition_function(
         self,
@@ -387,38 +437,129 @@ class ProbabilityOfFeasibility(SingleModelAcquisitionBuilder[ProbabilisticModel]
         return function  # no need to update anything
 
 
-def probability_of_feasibility(
-    model: ProbabilisticModel, threshold: float | TensorType
-) -> AcquisitionFunction:
-    r"""
-    The probability of feasibility acquisition function defined in :cite:`gardner14` as
-
-    .. math::
-
-        \int_{-\infty}^{\tau} p(c(\mathbf{x}) | \mathbf{x}, \mathcal{D}) \mathrm{d} c(\mathbf{x})
-        \qquad ,
-
-    where :math:`\tau` is a threshold. Values below the threshold are considered feasible by the
-    constraint function.
-
-    :param model: The model of the objective function.
-    :param threshold: The (scalar) probability of feasibility threshold.
-    :return: The probability of feasibility function. This function will raise
+class probability_below_threshold(AcquisitionFunctionClass):
+    def __init__(self, model: ProbabilisticModel, threshold: float | TensorType):
+        r"""
+        The probability of being below the threshold. This brings together commonality
+        between probability of improvement and probability of feasiblity.
+        Probability is is caculated with respect to the `model` posterior.
+        For model posterior :math:`f`, this is
+        .. math:: x \mapsto \mathbb P \left (f(x) < \eta)\right]
+        where :math:`\eta` is the threshold.
+        :param model: The model of the objective function.
+        :param threshold: The (scalar) probability of feasibility threshold.
+        :return: The probability of feasibility function. This function will raise
         :exc:`ValueError` or :exc:`~tf.errors.InvalidArgumentError` if used with a batch size
         greater than one.
-    :raise ValueError or tf.errors.InvalidArgumentError: If ``threshold`` is not a scalar.
-    """
-    tf.debugging.assert_scalar(threshold)
+        :raise ValueError or tf.errors.InvalidArgumentError: If ``threshold`` is not a scalar.
+        """
+        tf.debugging.assert_scalar(threshold)
+        self._model = model
+        self._threshold = tf.Variable(threshold)
 
     @tf.function
-    def acquisition(x: TensorType) -> TensorType:
+    def __call__(self, x: TensorType) -> TensorType:
         tf.debugging.assert_shapes(
             [(x, [..., 1, None])],
             message="This acquisition function only supports batch sizes of one.",
         )
-        mean, var = model.predict(tf.squeeze(x, -2))
+        mean, var = self._model.predict(tf.squeeze(x, -2))
         distr = tfp.distributions.Normal(mean, tf.sqrt(var))
-        return distr.cdf(tf.cast(threshold, x.dtype))
+        return distr.cdf(tf.cast(self._threshold, x.dtype))
+
+    def update(self, threshold: TensorType) -> None:
+        """Update the acquisition function with a new threshold value."""
+        self._threshold.assign(threshold)
+
+
+class FastConstraintsFeasibility(SingleModelAcquisitionBuilder[ProbabilisticModel]):
+    """
+    Builds a feasiblity acquisition function from the residuals of explicit constraints defined in
+    the search space.
+    """
+
+    def __init__(
+        self,
+        search_space: SearchSpace,
+        smoothing_function: Optional[Callable[[TensorType], TensorType]] = None,
+    ):
+        """
+        :param search_space: The global search space over which the feasibility of the constraints
+            is defined.
+        :param smoothing_function: The smoothing function used for constraints residuals. The
+            default is CDF of the Normal distribution with a scale of `1e-3`.
+        :raise NotImplementedError: If the `search_space` does not have constraints.
+        """
+        if not search_space.has_constraints:
+            raise NotImplementedError(
+                "FastConstraintsFeasibility requires constraints in the search space."
+            )
+
+        self._search_space = search_space
+        self._smoothing_function = smoothing_function
+
+    def __repr__(self) -> str:
+        """"""
+        return f"FastConstraintsFeasibility({self._search_space!r}, {self._smoothing_function!r})"
+
+    def prepare_acquisition_function(
+        self,
+        model: ProbabilisticModel,
+        dataset: Optional[Dataset] = None,
+    ) -> AcquisitionFunction:
+        """
+        :param model: Unused.
+        :param dataset: Unused.
+        :return: The function for feasibility of constraints.
+        """
+        return fast_constraints_feasibility(self._search_space, self._smoothing_function)
+
+    def update_acquisition_function(
+        self,
+        function: AcquisitionFunction,
+        model: ProbabilisticModel,
+        dataset: Optional[Dataset] = None,
+    ) -> AcquisitionFunction:
+        """
+        :param function: The acquisition function to update.
+        :param model: Unused.
+        :param dataset: Unused.
+        :return: The function for feasibility of constraints.
+        """
+        return function  # No need to update anything.
+
+
+def fast_constraints_feasibility(
+    search_space: SearchSpace,
+    smoothing_function: Optional[Callable[[TensorType], TensorType]] = None,
+) -> AcquisitionFunction:
+    """
+    Returns a feasiblity acquisition function from the residuals of explicit constraints defined in
+    the search space.
+
+    :param search_space: The global search space over which the feasibility of the constraints
+        is defined.
+    :param smoothing_function: The smoothing function used for constraints residuals. The
+        default is CDF of the Normal distribution with a scale of `1e-3`.
+    :return: The function for feasibility of constraints.
+    :raise NotImplementedError: If the `search_space` does not have constraints.
+    """
+
+    if not search_space.has_constraints:
+        raise NotImplementedError(
+            "fast_constraints_feasibility requires constraints in the search space."
+        )
+
+    @tf.function
+    def acquisition(x: TensorType) -> TensorType:
+        if smoothing_function is None:
+            _smoothing_function = tfp.distributions.Normal(
+                tf.cast(0.0, x.dtype), tf.cast(1e-3, x.dtype)
+            ).cdf
+        else:
+            _smoothing_function = smoothing_function
+        residuals = search_space.constraints_residuals(x)
+        return tf.math.reduce_prod(_smoothing_function(residuals), axis=-1)
 
     return acquisition
 
@@ -433,7 +574,7 @@ class ExpectedConstrainedImprovement(AcquisitionFunctionBuilder[ProbabilisticMod
 
     def __init__(
         self,
-        objective_tag: str,
+        objective_tag: Tag,
         constraint_builder: AcquisitionFunctionBuilder[ProbabilisticModelType],
         min_feasibility_probability: float | TensorType = 0.5,
     ):
@@ -471,8 +612,8 @@ class ExpectedConstrainedImprovement(AcquisitionFunctionBuilder[ProbabilisticMod
 
     def prepare_acquisition_function(
         self,
-        models: Mapping[str, ProbabilisticModelType],
-        datasets: Optional[Mapping[str, Dataset]] = None,
+        models: Mapping[Tag, ProbabilisticModelType],
+        datasets: Optional[Mapping[Tag, Dataset]] = None,
     ) -> AcquisitionFunction:
         """
         :param models: The models over each tag.
@@ -484,7 +625,7 @@ class ExpectedConstrainedImprovement(AcquisitionFunctionBuilder[ProbabilisticMod
         :raise tf.errors.InvalidArgumentError: If the objective data is empty.
         """
         tf.debugging.Assert(datasets is not None, [tf.constant([])])
-        datasets = cast(Mapping[str, Dataset], datasets)
+        datasets = cast(Mapping[Tag, Dataset], datasets)
 
         objective_model = models[self._objective_tag]
         objective_dataset = datasets[self._objective_tag]
@@ -520,8 +661,8 @@ class ExpectedConstrainedImprovement(AcquisitionFunctionBuilder[ProbabilisticMod
     def update_acquisition_function(
         self,
         function: AcquisitionFunction,
-        models: Mapping[str, ProbabilisticModelType],
-        datasets: Optional[Mapping[str, Dataset]] = None,
+        models: Mapping[Tag, ProbabilisticModelType],
+        datasets: Optional[Mapping[Tag, Dataset]] = None,
     ) -> AcquisitionFunction:
         """
         :param function: The acquisition function to update.
@@ -529,7 +670,7 @@ class ExpectedConstrainedImprovement(AcquisitionFunctionBuilder[ProbabilisticMod
         :param datasets: The data from the observer.
         """
         tf.debugging.Assert(datasets is not None, [tf.constant([])])
-        datasets = cast(Mapping[str, Dataset], datasets)
+        datasets = cast(Mapping[Tag, Dataset], datasets)
 
         objective_model = models[self._objective_tag]
         objective_dataset = datasets[self._objective_tag]
@@ -956,7 +1097,6 @@ class batch_monte_carlo_expected_improvement(AcquisitionFunctionClass):
         """
         :param sample_size: The number of Monte-Carlo samples.
         :param model: The model of the objective function.
-        :param sampler:  ReparametrizationSampler.
         :param eta: The "best" observation.
         :param jitter: The size of the jitter to use when stabilising the Cholesky decomposition of
             the covariance matrix.
@@ -991,6 +1131,621 @@ class batch_monte_carlo_expected_improvement(AcquisitionFunctionClass):
         return tf.reduce_mean(batch_improvement, axis=-1, keepdims=True)  # [..., 1]
 
 
+class BatchExpectedImprovement(SingleModelAcquisitionBuilder[ProbabilisticModel]):
+    """Accurate approximation of the batch expected improvement, using the
+    method of Chvallier and Ginsbourger :cite:`chevalier2013fast`.
+
+    Internally, this uses a highly accurate approximation of the cumulative
+    density function of the multivariate Gaussian, developed by Alan Genz
+    :cite:`genz2016numerical`.
+    """
+
+    def __init__(
+        self,
+        sample_size: int,
+        *,
+        jitter: float = DEFAULTS.JITTER,
+    ):
+        """Initialise the BatchExpectedImprovement instance.
+
+        :param sample_size: int, number of Sobol samples to use.
+        :param jitter: float, amount of jitter for Cholesky factorisations.
+        """
+
+        tf.debugging.assert_positive(sample_size)
+        tf.debugging.assert_greater_equal(jitter, 0.0)
+
+        self._sample_size = sample_size
+        self._jitter = jitter
+
+    def __repr__(self) -> str:
+        """"""
+
+        return f"BatchExpectedImprovement({self._sample_size!r}, " f"jitter={self._jitter!r})"
+
+    def prepare_acquisition_function(
+        self,
+        model: ProbabilisticModel,
+        dataset: Optional[Dataset] = None,
+    ) -> AcquisitionFunction:
+        """
+        :param model: The model. Must have event shape [1].
+        :param dataset: The data from the observer. Must be populated.
+        :return: The batch *expected improvement* acquisition function.
+        :raise ValueError (or InvalidArgumentError): If ``dataset`` is not populated, or ``model``
+            does not have an event shape of [1].
+        """
+        tf.debugging.Assert(dataset is not None, [])
+        dataset = cast(Dataset, dataset)
+        tf.debugging.assert_positive(len(dataset), message="Dataset must be populated.")
+
+        # Get mean and covariance
+        mean, _ = model.predict(dataset.query_points)
+
+        tf.debugging.assert_shapes(
+            [(mean, ["_", 1])],
+            message="Expected model with event shape [1].",
+        )
+
+        eta = tf.reduce_min(mean, axis=0)
+
+        acquisition_function = batch_expected_improvement(
+            self._sample_size,
+            model,
+            eta,
+            self._jitter,
+        )
+
+        return acquisition_function
+
+    def update_acquisition_function(
+        self,
+        function: AcquisitionFunction,
+        model: ProbabilisticModel,
+        dataset: Optional[Dataset] = None,
+    ) -> AcquisitionFunction:
+        """
+        :param function: The acquisition function to update.
+        :param model: The model. Must have event shape [1].
+        :param dataset: The data from the observer. Must be populated.
+        """
+        tf.debugging.Assert(dataset is not None, [])
+        dataset = cast(Dataset, dataset)
+        tf.debugging.assert_positive(len(dataset), message="Dataset must be populated.")
+        tf.debugging.Assert(isinstance(function, batch_expected_improvement), [])
+
+        # Get mean and covariance
+        mean, _ = model.predict(dataset.query_points)
+        eta = tf.reduce_min(mean, axis=0)
+
+        function.update(eta=eta)  # type: ignore
+
+        return function
+
+
+class batch_expected_improvement(AcquisitionFunctionClass):
+    def __init__(
+        self,
+        sample_size: int,
+        model: ProbabilisticModel,
+        eta: TensorType,
+        jitter: float,
+    ):
+        """Initialise the batch_expected_improvement instance.
+
+        :param sample_size: int, number of samples to use.
+        :param model: Gaussian process regression model.
+        :param eta: Tensor of shape (,), expected improvement threshold. This
+            is the best value observed so far durin the BO loop.
+        :param jitter: float, amount of jitter for Cholesky factorisations.
+        """
+
+        self._sample_size = sample_size
+        self._jitter = jitter
+        self._eta = tf.Variable(eta)
+        self._model = model
+
+        self._mvn_cdf_1: Optional[MultivariateNormalCDF] = None
+        self._mvn_cdf_2: Optional[MultivariateNormalCDF] = None
+
+        self._num_sobol_skip = int(tf.math.floor(10 ** 9 * tf.random.uniform((), dtype=tf.float32)))
+
+    def update(self, eta: TensorType) -> None:
+        """Update the acquisition function with a new eta value and reset the
+        reparam sampler.
+        """
+        self._eta.assign(eta)
+        self._num_sobol_skip = int(tf.math.floor(10 ** 9 * tf.random.uniform((), dtype=tf.float32)))
+
+    def _compute_bm(
+        self,
+        mean: tf.Tensor,
+        threshold: tf.Tensor,
+    ) -> TensorType:
+        """Helper function for the batch expected improvement, which computes
+        the tensors b and m as detailed in Chevalier and Ginsbourger
+        :cite:`chevalier2013fast`.
+
+        :param mean: Tensor of shape (B, Q)
+        :param threshold: Tensor of shape (B,)
+
+        :returns b: Tensor of shape (B, Q, Q)
+        :returns m: Tensor of shape (B, Q, Q)
+        """
+
+        # Check shapes of input tensors
+        tf.debugging.assert_shapes(
+            [
+                (mean, ("B", "Q")),
+                (threshold, ("B",)),
+            ]
+        )
+
+        # Unpack tensor shape and data type
+        B, Q = mean.shape
+        dtype = mean.dtype
+
+        # Compute b tensor
+        threshold = tf.tile(threshold[:, None], (1, Q))
+        threshold = tf.linalg.diag(threshold)  # (B, Q, Q)
+
+        b = tf.zeros(shape=(B, Q, Q), dtype=dtype)
+        b = b - threshold
+
+        # Compute m tensor
+        m = mean[:, None, :] - mean[:, :, None]  # (B, Q, Q)
+        m = m - tf.linalg.diag(mean)  # (B, Q, Q)
+
+        return b, m
+
+    def _delta(self, idx: int, dim: int, B: int, transpose: bool, dtype: tf.DType) -> TensorType:
+        """Helper function for the _compute_Sigma function, which computes a
+        *delta* tensor of shape (B, idx, idx) such that
+
+            delta[B, i, :] = 1 if i == idx
+            delta[B, i, :] = 0 otherwise.
+
+        If transpose == True, then the last two dimensions of the tensor are
+        transposed, in which case
+
+            delta[B, :, i] = 1 if i == idx
+            delta[B, :, i] = 0 otherwise.
+
+        :param idx: Index for entries equal to 1.
+        :param dim: Dimension of the last and second to last axes.
+        :param B: Leading dimension of tensor.
+        :param transpose: Whether to transpose the last two dimensions or not.
+        :param dtype: The dtype of the tensor, either tf.float32 or tf.float64.
+        """
+
+        # Check input parameters
+        tf.debugging.assert_non_negative(idx)
+        tf.debugging.assert_non_negative(dim)
+        tf.debugging.assert_positive(B)
+
+        o1 = tf.ones(shape=(B, idx, dim), dtype=dtype)
+        z1 = tf.zeros(shape=(B, 1, dim), dtype=dtype)
+        o2 = tf.ones(shape=(B, dim - idx - 1, dim), dtype=dtype)
+
+        delta = tf.concat([o1, z1, o2], axis=1)
+        delta = tf.transpose(delta, perm=[0, 2, 1]) if transpose else delta
+
+        return delta
+
+    def _compute_Sigma(
+        self,
+        covariance: tf.Tensor,
+    ) -> TensorType:
+        """Helper function for the batch expected improvement, which computes
+        the tensor Sigma, as detailed in Chevalier and Ginsbourger
+        :cite:`chevalier2013fast`.
+
+        :param covariance: Tensor of shape (B, Q, Q)
+        :returns Sigma: Tensor of shape (B, Q, Q, Q)
+        """
+
+        # Check shapes of covariance tensor
+        tf.debugging.assert_shapes([(covariance, ("B", "Q", "Q"))])
+
+        # Unpack tensor shape and dtype
+        B, Q, _ = covariance.shape
+        dtype = covariance.dtype
+
+        Sigma = tf.zeros(shape=(B, Q, Q, Q))
+
+        def compute_single_slice(q: int) -> TensorType:
+
+            diq = self._delta(q, Q, B, transpose=False, dtype=dtype)
+            dqj = self._delta(q, Q, B, transpose=True, dtype=dtype)
+
+            Sigma_ij = covariance[:, :, :]
+            Sigma_iq = covariance[:, :, q : q + 1]
+            Sigma_qj = covariance[:, q : q + 1, :]
+            Sigma_qq = covariance[:, q : q + 1, q : q + 1]
+
+            cov = Sigma_ij * diq * dqj - Sigma_iq * diq - Sigma_qj * dqj + Sigma_qq
+
+            return cov
+
+        Sigma = tf.map_fn(
+            compute_single_slice,
+            tf.range(Q),
+            fn_output_signature=dtype,
+        )
+
+        Sigma = tf.transpose(Sigma, perm=[1, 0, 2, 3])
+
+        return Sigma
+
+    def _compute_p(
+        self,
+        m_reshaped: tf.Tensor,
+        b_reshaped: tf.Tensor,
+        Sigma_reshaped: tf.Tensor,
+        mvn_cdf: Callable[[TensorType, TensorType, TensorType, float], TensorType],
+    ) -> TensorType:
+        """Helper function for the batch expected improvement, which computes
+        the tensor p, as detailed in Chevalier and Ginsbourger
+        :cite:`chevalier2013fast`.
+
+        :param m_reshaped: Tensor of shape (BQ, Q)
+        :param b_reshaped: Tensor of shape (BQ, Q)
+        :param Sigma_reshaped: Tensor of shape (BQ, Q, Q)
+        :returns p: Tensor of shape (B, Q)
+        """
+
+        # Check shapes of covariance tensor
+        tf.debugging.assert_shapes(
+            [
+                (m_reshaped, ("BQ", "Q")),
+                (b_reshaped, ("BQ", "Q")),
+                (Sigma_reshaped, ("BQ", "Q", "Q")),
+            ]
+        )
+
+        # Unpack dtype and mean shape
+        dtype = m_reshaped.dtype
+        BQ, Q = m_reshaped.shape  # (B*Q, Q)
+
+        if BQ % Q == 0:
+            B = BQ // Q
+
+        else:
+            raise ValueError(
+                f"Expected size of dimension 0 of m_reshaped tensor to be "
+                f"divisible by size of dimension 1, instead found "
+                f"{m_reshaped.shape[0]} and {m_reshaped.shape[1]}."
+            )
+
+        # Compute mean, covariance and x for p mvn normal cdf
+        p_cdf_mean = tf.zeros(shape=(BQ, Q), dtype=dtype)  # (B*Q, Q)
+        p_cdf_cov = Sigma_reshaped  # (B*Q, Q, Q)
+
+        p_cdf_x = b_reshaped - m_reshaped  # (B*Q, Q)
+
+        p = mvn_cdf(  # type: ignore
+            x=p_cdf_x,
+            mean=p_cdf_mean,
+            cov=p_cdf_cov,
+        )  # (B*Q,)
+
+        p = tf.reshape(p, shape=(B, Q))  # (B, Q)
+
+        return p
+
+    def _compute_c(
+        self,
+        m_reshaped: tf.Tensor,
+        b_reshaped: tf.Tensor,
+        Sigma_reshaped: tf.Tensor,
+    ) -> TensorType:
+        """Helper function for the batch expected improvement, which computes
+        the tensor c, which is the c^{(i)} tensor detailed in Chevalier and
+        Ginsbourger :cite:`chevalier2013fast`.
+
+        :param m_reshaped: Tensor of shape (BQ, Q)
+        :param b_reshaped: Tensor of shape (BQ, Q)
+        :param Sigma_reshaped: Tensor of shape (BQ, Q, Q)
+        :returns c: Tensor of shape (B, Q, Q-1)
+        """
+
+        # Check shapes of covariance tensor
+        tf.debugging.assert_shapes(
+            [
+                (m_reshaped, ("BQ", "Q")),
+                (b_reshaped, ("BQ", "Q")),
+                (Sigma_reshaped, ("BQ", "Q", "Q")),
+            ]
+        )
+
+        # Unpack tensor shape
+        BQ, Q = m_reshaped.shape
+
+        # Compute difference between b and m tensors
+        diff = b_reshaped - m_reshaped  # (B*Q, Q)
+
+        # Compute c, including the ith entry, which we want to remove
+        cov_ratio = Sigma_reshaped / tf.linalg.diag_part(Sigma_reshaped)[:, :, None]  # (B*Q, Q, Q)
+        c = diff[:, None, :] - diff[:, :, None] * cov_ratio  # (B*Q, Q, Q)
+
+        # Remove the ith entry by masking c with a boolean mask with False across
+        # the diagonal and True in the off-diagonal terms
+        mask = tf.math.logical_not(tf.cast(tf.eye(Q, dtype=tf.int32), dtype=tf.bool))
+        mask = tf.tile(mask[None, :, :], (c.shape[0], 1, 1))
+
+        c = tf.ragged.boolean_mask(c, mask).to_tensor()
+
+        return c
+
+    def _compute_R(
+        self,
+        Sigma_reshaped: tf.Tensor,
+    ) -> TensorType:
+        """Helper function for the batch expected improvement, which computes
+        the tensor R, which is the Sigma^{(i)} tensor detailed in Chevalier
+        and Ginsbourger :cite:`chevalier2013fast`.
+
+        :param Sigma_reshaped: Tensor of shape (BQ, Q, Q)
+        :returns R: Tensor of shape (B, Q-1, Q-1)
+        """
+
+        # Check shapes of covariance tensor
+        tf.debugging.assert_shapes([(Sigma_reshaped, ("BQ", "Q", "Q"))])
+
+        # Unpack tensor shape
+        BQ, Q, _ = Sigma_reshaped.shape
+
+        Sigma_uv = tf.tile(Sigma_reshaped[:, None, :, :], (1, Q, 1, 1))
+        Sigma_iu = tf.tile(Sigma_reshaped[:, :, :, None], (1, 1, 1, Q))
+        Sigma_iv = tf.tile(Sigma_reshaped[:, :, None, :], (1, 1, Q, 1))
+        Sigma_ii = tf.linalg.diag_part(Sigma_reshaped)[:, :, None, None]
+
+        R_whole = Sigma_uv - Sigma_iu * Sigma_iv / Sigma_ii
+
+        def create_blocks(q: int) -> TensorType:
+
+            block1 = tf.concat(
+                [
+                    R_whole[:, q, :q, :q],
+                    R_whole[:, q, q + 1 :, :q],
+                ],
+                axis=1,
+            )
+
+            block2 = tf.concat(
+                [
+                    R_whole[:, q, :q, q + 1 :],
+                    R_whole[:, q, q + 1 :, q + 1 :],
+                ],
+                axis=1,
+            )
+
+            R_block = tf.concat([block1, block2], axis=2)
+
+            return R_block
+
+        R = tf.map_fn(
+            create_blocks,
+            tf.range(Q),
+            fn_output_signature=R_whole.dtype,
+        )
+        R = tf.transpose(R, perm=[1, 0, 2, 3])
+
+        return R
+
+    def _compute_Phi(
+        self,
+        c: tf.Tensor,
+        R: tf.Tensor,
+        mvn_cdf: Callable[[TensorType, TensorType, TensorType, float], TensorType],
+    ) -> TensorType:
+        """Helper function for the batch expected improvement, which computes
+        the tensor Phi, which is the tensor of multivariate Gaussian CDFs, in
+        the inner sum of the equation (3) in Chevalier and Ginsbourger
+        :cite:`chevalier2013fast`.
+
+        :param c: Tensor of shape (BQ, Q, Q-1).
+        :param R: Tensor of shape (BQ, Q, Q-1, Q-1).
+        :param mvn_cdf: Multivariate Gaussian CDF, made using MultivariateNormalCDF.
+        :returns Phi: Tensor of multivariate Gaussian CDFs.
+        """
+
+        # Check shapes of covariance tensor
+        tf.debugging.assert_shapes(
+            [
+                (c, ("BQ", "Q", "Q_")),
+                (R, ("BQ", "Q", "Q_", "Q_")),
+            ]
+        )
+
+        # Unpack tensor shape and data type
+        BQ, Q, _, Q_ = R.shape
+        dtype = R.dtype
+
+        try:
+            assert BQ % Q == 0
+
+        except AssertionError:
+            raise ValueError(
+                f"Expected size of dimension 0 of R tensor to be "
+                f"divisible by size of dimension 1, instead found "
+                f"{R.shape[0]} and {R.shape[1]}."
+            )
+
+        # Compute parallelisation dimension from batch size
+        B = BQ // Q
+
+        c_reshaped = tf.reshape(c, (BQ * Q, Q - 1))
+        R_reshaped = tf.reshape(R, (BQ * Q, Q - 1, Q - 1))
+
+        # Compute mean, covariance and x for Phi mvn normal cdf
+        Phi_cdf_x = c_reshaped  # (B*Q, Q-1)
+        Phi_cdf_mean = tf.zeros(shape=(BQ * Q, Q - 1), dtype=dtype)  # (B*Q*Q, Q)
+        Phi_cdf_cov = R_reshaped  # (B*Q*Q, Q-1, Q-1)
+
+        # Compute multivariate cdfs
+        mvn_cdfs = mvn_cdf(  # type: ignore
+            x=Phi_cdf_x,
+            mean=Phi_cdf_mean,
+            cov=Phi_cdf_cov,
+        )
+        mvn_cdfs = tf.reshape(mvn_cdfs, (B, Q, Q))  # (B, Q, Q)
+
+        return mvn_cdfs
+
+    def _compute_batch_expected_improvement(
+        self,
+        mean: tf.Tensor,
+        covariance: tf.Tensor,
+        threshold: tf.Tensor,
+        mvn_cdf_1: Callable[[TensorType, TensorType, TensorType, float], TensorType],
+        mvn_cdf_2: Callable[[TensorType, TensorType, TensorType, float], TensorType],
+    ) -> TensorType:
+        """Accurate Monte Carlo approximation of the batch expected
+        improvement, using the method of Chevalier and Ginsbourger
+        :cite:`chevalier2013fast`.
+
+        :param mean: Tensor of shape (B, Q).
+        :param covariance: Tensor of shape (B, Q, Q).
+        :param threshold: Tensor of shape (B, Q).
+        :param mvn_cdf_1: Callable computing the multivariate CDF of a Q-dimensional Gaussian.
+        :param mvn_cdf_2: Callable computing the multivariate CDF of a (Q-1)-dimensional Gaussian.
+        :returns ei: Tensor of shape (B,), expected improvement.
+        """
+
+        # Check shapes of covariance tensor
+        tf.debugging.assert_shapes(
+            [
+                (mean, ("B", "Q")),
+                (covariance, ("B", "Q", "Q")),
+                (threshold, ("B",)),
+            ]
+        )
+
+        # Unpack and mean shape
+        B, Q = mean.shape
+
+        # Compute b and m tensors
+        b, m = self._compute_bm(
+            mean=mean,
+            threshold=threshold,
+        )  # (B, Q, Q), (B, Q, Q)
+
+        # Compute Sigma
+        Sigma = self._compute_Sigma(covariance=covariance)  # (B, Q, Q, Q)
+
+        # Reshape all tensors, for batching
+        b_reshaped = tf.reshape(b, (B * Q, Q))
+        m_reshaped = tf.reshape(m, (B * Q, Q))
+        Sigma_reshaped = tf.reshape(Sigma, (B * Q, Q, Q))
+
+        # Compute p tensor
+        p = self._compute_p(
+            m_reshaped=m_reshaped,
+            b_reshaped=b_reshaped,
+            Sigma_reshaped=Sigma_reshaped,
+            mvn_cdf=mvn_cdf_1,
+        )
+
+        # Compute c
+        c = self._compute_c(
+            m_reshaped=m_reshaped,
+            b_reshaped=b_reshaped,
+            Sigma_reshaped=Sigma_reshaped,
+        )  # (B*Q, Q, Q-1)
+
+        # Compute Sigma_i
+        R = self._compute_R(
+            Sigma_reshaped=Sigma_reshaped,
+        )  # (B*Q, Q, Q-1, Q-1)
+
+        # Compute Q-1 multivariate CDFs
+        Phi_mvn_cdfs = self._compute_Phi(
+            c=c,
+            R=R,
+            mvn_cdf=mvn_cdf_2,
+        )
+
+        # Compute univariate pdfs
+        S_diag = tf.linalg.diag_part(Sigma)
+        normal = tfp.distributions.Normal(loc=m, scale=S_diag ** 0.5)
+        uvn_pdfs = tf.math.exp(normal.log_prob(b))  # (B, Q, Q)
+
+        Sigma_diag = tf.linalg.diag_part(tf.transpose(Sigma, perm=[0, 2, 1, 3]))
+        Sigma_diag = tf.transpose(Sigma_diag, perm=[0, 2, 1])
+
+        T = tf.tile(threshold[:, None], (1, Q))
+
+        mean_T_term = (mean - T) * p
+
+        # Compute inner sum
+        sum_term = tf.reduce_sum(
+            Sigma_diag * uvn_pdfs * Phi_mvn_cdfs,
+            axis=2,
+        )
+
+        # Compute outer sum
+        expected_improvement = tf.reduce_sum(mean_T_term + sum_term, axis=1)
+
+        return expected_improvement
+
+    @tf.function
+    def __call__(self, x: TensorType) -> TensorType:
+        """Computes the accurate approximation of the multi-point expected
+        improvement.
+
+        :param x: Tensor of shape (B, Q, D).
+        :returns ei: Tensor of shape (B,), expected improvement.
+        """
+
+        if self._mvn_cdf_1 is None:
+            self._mvn_cdf_1 = MultivariateNormalCDF(
+                sample_size=self._sample_size,
+                dim=x.shape[1],
+                dtype=x.dtype,
+                num_sobol_skip=self._num_sobol_skip,
+            )
+
+        if self._mvn_cdf_2 is None:
+            self._mvn_cdf_2 = MultivariateNormalCDF(
+                sample_size=self._sample_size,
+                dim=x.shape[1] - 1,
+                dtype=x.dtype,
+                num_sobol_skip=self._num_sobol_skip,
+            )
+
+        mean, covariance = self._model.predict_joint(x)  # type: ignore
+
+        mean = mean[:, :, 0]
+        covariance = covariance[:, 0, :, :]
+        covariance = covariance + 1e-6 * tf.eye(
+            covariance.shape[-1],
+            dtype=covariance.dtype,
+        )[None, :, :]
+
+        threshold = tf.tile(self._eta, (mean.shape[0],))
+
+        # Check shapes of x, mean, covariance and threshold tensors
+        tf.debugging.assert_shapes(
+            [
+                (x, ("B", "Q", "D")),
+                (mean, ("B", "Q")),
+                (covariance, ("B", "Q", "Q")),
+                (threshold, ("B",)),
+            ]
+        )
+
+        ei = self._compute_batch_expected_improvement(
+            mean=-mean,
+            covariance=covariance,
+            threshold=-threshold,
+            mvn_cdf_1=self._mvn_cdf_1,
+            mvn_cdf_2=self._mvn_cdf_2,
+        )[:, None]
+
+        return ei
+
+
 class MultipleOptimismNegativeLowerConfidenceBound(
     SingleModelVectorizedAcquisitionBuilder[ProbabilisticModel]
 ):
