@@ -19,7 +19,7 @@ acquisition functions.
 from __future__ import annotations
 
 from abc import ABC, abstractmethod
-from typing import Generic
+from typing import Callable, Generic
 
 import tensorflow as tf
 import tensorflow_probability as tfp
@@ -28,6 +28,7 @@ from scipy.optimize import bisect
 from ..models import ProbabilisticModel
 from ..models.interfaces import HasTrajectorySampler, ProbabilisticModelType
 from ..types import TensorType
+from .utils import select_nth_output
 
 
 class ThompsonSampler(ABC, Generic[ProbabilisticModelType]):
@@ -54,11 +55,21 @@ class ThompsonSampler(ABC, Generic[ProbabilisticModelType]):
         """
 
     @abstractmethod
-    def sample(self, model: ProbabilisticModelType, sample_size: int, at: TensorType) -> TensorType:
+    def sample(
+        self,
+        model: ProbabilisticModelType,
+        sample_size: int,
+        at: TensorType,
+        select_output: Callable[[TensorType], TensorType] = select_nth_output,
+    ) -> TensorType:
         """
         :param model: The model to sample from.
         :param sample_size: The desired number of samples.
         :param at: Input points that define the sampler.
+        :param select_output: A method that returns the desired output from the model sampler, with
+            shape `[S, N]` where `S` is the number of samples and `N` is the number of locations.
+            Defaults to the :func:~`trieste.acquisition.utils.select_nth_output` function with
+            output dimension 0.
         :return: Samples.
         """
 
@@ -72,23 +83,33 @@ class ExactThompsonSampler(ThompsonSampler[ProbabilisticModel]):
     with a sampling method.
     """
 
-    def sample(self, model: ProbabilisticModel, sample_size: int, at: TensorType) -> TensorType:
+    def sample(
+        self,
+        model: ProbabilisticModel,
+        sample_size: int,
+        at: TensorType,
+        select_output: Callable[[TensorType], TensorType] = select_nth_output,
+    ) -> TensorType:
         """
-        Return exact samples from either the objective function's minimser or its minimal value
+        Return exact samples from either the objective function's minimiser or its minimal value
         over the candidate set `at`.
 
         :param model: The model to sample from.
         :param sample_size: The desired number of samples.
         :param at: Where to sample the predictive distribution, with shape `[N, D]`, for points
             of dimension `D`.
+        :param select_output: A method that returns the desired output from the model sampler, with
+            shape `[S, N]` where `S` is the number of samples and `N` is the number of locations.
+            Defaults to the :func:~`trieste.acquisition.utils.select_nth_output` function with
+            output dimension 0.
         :return: The samples, of shape `[S, D]` (where `S` is the `sample_size`) if sampling
-            the function's minimser or shape `[S, 1]` if sampling the function's mimimal value.
+            the function's minimiser or shape `[S, 1]` if sampling the function's mimimal value.
         :raise ValueError: If ``at`` has an invalid shape or if ``sample_size`` is not positive.
         """
         tf.debugging.assert_positive(sample_size)
         tf.debugging.assert_shapes([(at, ["N", None])])
 
-        samples = model.sample(at, sample_size)  # [S, N, 1]
+        samples = select_output(model.sample(at, sample_size))[..., None]  # [S, N, 1]
 
         if self._sample_min_value:
             thompson_samples = tf.reduce_min(samples, axis=1)  # [S, 1]
@@ -130,7 +151,13 @@ class GumbelSampler(ThompsonSampler[ProbabilisticModel]):
 
         super().__init__(sample_min_value)
 
-    def sample(self, model: ProbabilisticModel, sample_size: int, at: TensorType) -> TensorType:
+    def sample(
+        self,
+        model: ProbabilisticModel,
+        sample_size: int,
+        at: TensorType,
+        select_output: Callable[[TensorType], TensorType] = select_nth_output,
+    ) -> TensorType:
         """
         Return approximate samples from of the objective function's minimum value.
 
@@ -138,6 +165,9 @@ class GumbelSampler(ThompsonSampler[ProbabilisticModel]):
         :param sample_size: The desired number of samples.
         :param at: Points at where to fit the Gumbel distribution, with shape `[N, D]`, for points
             of dimension `D`. We recommend scaling `N` with search space dimension.
+        :param select_output: A method that returns the desired output from the model sampler, with
+            shape `[S, N]` where `S` is the number of samples and `N` is the number of locations.
+            Currently unused.
         :return: The samples, of shape `[S, 1]`, where `S` is the `sample_size`.
         :raise ValueError: If ``at`` has an invalid shape or if ``sample_size`` is not positive.
         """
@@ -186,7 +216,13 @@ class ThompsonSamplerFromTrajectory(ThompsonSampler[HasTrajectorySampler]):
     probabilistic model with a :meth:`trajectory_sampler` method.
     """
 
-    def sample(self, model: HasTrajectorySampler, sample_size: int, at: TensorType) -> TensorType:
+    def sample(
+        self,
+        model: ProbabilisticModel,
+        sample_size: int,
+        at: TensorType,
+        select_output: Callable[[TensorType], TensorType] = select_nth_output,
+    ) -> TensorType:
         """
         Return approximate samples from either the objective function's minimser or its minimal
         value over the candidate set `at`.
@@ -195,6 +231,10 @@ class ThompsonSamplerFromTrajectory(ThompsonSampler[HasTrajectorySampler]):
         :param sample_size: The desired number of samples.
         :param at: Where to sample the predictive distribution, with shape `[N, D]`, for points
             of dimension `D`.
+        :param select_output: A method that returns the desired output from the model sampler, with
+            shape `[S, N]` where `S` is the number of samples and `N` is the number of locations.
+            Defaults to the :func:~`trieste.acquisition.utils.select_nth_output` function with
+            output dimension 0.
         :return: The samples, of shape `[S, D]` (where `S` is the `sample_size`) if sampling
             the function's minimser or shape `[S, 1]` if sampling the function's mimimal value.
         :raise ValueError: If ``at`` has an invalid shape or if ``sample_size`` is not positive.
@@ -218,7 +258,7 @@ class ThompsonSamplerFromTrajectory(ThompsonSampler[HasTrajectorySampler]):
         for _ in tf.range(sample_size):
             sampled_trajectory = trajectory_sampler.get_trajectory()
             expanded_at = tf.expand_dims(at, -2)  # [N, 1, D]
-            evaluated_trajectory = sampled_trajectory(expanded_at)  # [N, 1]
+            evaluated_trajectory = select_output(sampled_trajectory(expanded_at))  # [N, 1]
             if self._sample_min_value:
                 sample = tf.reduce_min(evaluated_trajectory, keepdims=True)  # [1, 1]
             else:
